import { existsSync } from 'fs';
import mkdirp from 'mkdirp';
import * as path from 'path';
import { URI } from 'vscode-uri';
import { parseDocument } from './parseDocument';
import { tmpdir } from 'os';
import { WorkspaceMessageProcessor } from './WorkspaceMessageProcessor';
export class MessageProcessor {
    constructor({ logger, fileExtensions, graphqlFileExtensions, loadConfigOptions, config, parser, tmpDir, connection, }) {
        this._rootPath = process.cwd();
        this._sortedWorkspaceUris = [];
        this._processors = new Map();
        this._connection = connection;
        this._willShutdown = false;
        this._logger = logger;
        this._graphQLConfig = config;
        this._parser = (text, uri) => {
            const p = parser !== null && parser !== void 0 ? parser : parseDocument;
            return p(text, uri, fileExtensions, graphqlFileExtensions, this._logger);
        };
        this._tmpDir = tmpDir || tmpdir();
        const tmpDirBase = path.join(this._tmpDir, 'graphql-language-service');
        this._loadConfigOptions = { legacy: true, ...loadConfigOptions };
        if (!existsSync(tmpDirBase)) {
            mkdirp(tmpDirBase);
        }
    }
    get connection() {
        return this._connection;
    }
    set connection(connection) {
        this._connection = connection;
    }
    async handleInitializeRequest(params, _token, configDir) {
        var _a, _b;
        if (!params) {
            throw new Error('`params` argument is required to initialize.');
        }
        const serverCapabilities = {
            capabilities: {
                workspaceSymbolProvider: true,
                documentSymbolProvider: true,
                completionProvider: {
                    resolveProvider: true,
                    triggerCharacters: [' ', ':', '$', '(', '@'],
                },
                definitionProvider: true,
                textDocumentSync: 1,
                hoverProvider: true,
                workspace: {
                    workspaceFolders: {
                        supported: true,
                        changeNotifications: true,
                    },
                },
            },
        };
        this._sortedWorkspaceUris = (_b = (_a = params.workspaceFolders) === null || _a === void 0 ? void 0 : _a.map(ws => ws.uri).sort((a, b) => b.length - a.length)) !== null && _b !== void 0 ? _b : [
            URI.file(configDir ? configDir.trim() : params.rootUri || this._rootPath).toString(),
        ];
        this._sortedWorkspaceUris.forEach(uri => {
            this._processors.set(uri, new WorkspaceMessageProcessor({
                connection: this._connection,
                loadConfigOptions: this._loadConfigOptions,
                logger: this._logger,
                parser: this._parser,
                tmpDir: this._tmpDir,
                config: this._graphQLConfig,
                rootPath: URI.parse(uri).fsPath,
            }));
        });
        if (!serverCapabilities) {
            throw new Error('GraphQL Language Server is not initialized.');
        }
        this._logger.info(JSON.stringify({
            type: 'usage',
            messageType: 'initialize',
        }));
        return serverCapabilities;
    }
    _findWorkspaceProcessor(uri) {
        const workspace = this._sortedWorkspaceUris.find(wsUri => uri.startsWith(wsUri));
        return this._processors.get(workspace !== null && workspace !== void 0 ? workspace : '');
    }
    async handleDidOpenOrSaveNotification(params) {
        var _a, _b;
        if (!params || !params.textDocument) {
            throw new Error('`textDocument` argument is required.');
        }
        const { uri } = params.textDocument;
        return ((_b = (_a = this._findWorkspaceProcessor(uri)) === null || _a === void 0 ? void 0 : _a.handleDidOpenOrSaveNotification(params)) !== null && _b !== void 0 ? _b : { uri, diagnostics: [] });
    }
    async handleDidChangeNotification(params) {
        var _a, _b;
        if (!params ||
            !params.textDocument ||
            !params.contentChanges ||
            !params.textDocument.uri) {
            throw new Error('`textDocument`, `textDocument.uri`, and `contentChanges` arguments are required.');
        }
        return ((_b = (_a = this._findWorkspaceProcessor(params.textDocument.uri)) === null || _a === void 0 ? void 0 : _a.handleDidChangeNotification(params)) !== null && _b !== void 0 ? _b : null);
    }
    async handleDidChangeConfiguration(_params) {
        await Promise.all(Array.from(this._processors.values()).map(processor => processor._updateGraphQLConfig()));
        this._logger.log(JSON.stringify({
            type: 'usage',
            messageType: 'workspace/didChangeConfiguration',
        }));
        return {};
    }
    handleDidCloseNotification(params) {
        var _a;
        if (!params || !params.textDocument) {
            throw new Error('`textDocument` is required.');
        }
        const { uri } = params.textDocument;
        (_a = this._findWorkspaceProcessor(uri)) === null || _a === void 0 ? void 0 : _a.handleDidCloseNotification(params);
    }
    handleShutdownRequest() {
        this._willShutdown = true;
        return;
    }
    handleExitNotification() {
        process.exit(this._willShutdown ? 0 : 1);
    }
    validateDocumentAndPosition(params) {
        if (!params ||
            !params.textDocument ||
            !params.textDocument.uri ||
            !params.position) {
            throw new Error('`textDocument`, `textDocument.uri`, and `position` arguments are required.');
        }
    }
    async handleCompletionRequest(params) {
        var _a, _b;
        this.validateDocumentAndPosition(params);
        return ((_b = (_a = this._findWorkspaceProcessor(params.textDocument.uri)) === null || _a === void 0 ? void 0 : _a.handleCompletionRequest(params)) !== null && _b !== void 0 ? _b : []);
    }
    async handleHoverRequest(params) {
        var _a, _b;
        this.validateDocumentAndPosition(params);
        return ((_b = (_a = this._findWorkspaceProcessor(params.textDocument.uri)) === null || _a === void 0 ? void 0 : _a.handleHoverRequest(params)) !== null && _b !== void 0 ? _b : {
            contents: [],
        });
    }
    async handleWatchedFilesChangedNotification(params) {
        return Promise.all(params.changes.map(async (change) => {
            var _a, _b;
            return ((_b = (_a = this._findWorkspaceProcessor(change.uri)) === null || _a === void 0 ? void 0 : _a.handleWatchedFileChangedNotification(change)) !== null && _b !== void 0 ? _b : undefined);
        }));
    }
    async handleDefinitionRequest(params, _token) {
        var _a, _b;
        if (!params || !params.textDocument || !params.position) {
            throw new Error('`textDocument` and `position` arguments are required.');
        }
        return ((_b = (_a = this._findWorkspaceProcessor(params.textDocument.uri)) === null || _a === void 0 ? void 0 : _a.handleDefinitionRequest(params, _token)) !== null && _b !== void 0 ? _b : []);
    }
    async handleDocumentSymbolRequest(params) {
        var _a, _b;
        if (!params || !params.textDocument) {
            throw new Error('`textDocument` argument is required.');
        }
        return ((_b = (_a = this._findWorkspaceProcessor(params.textDocument.uri)) === null || _a === void 0 ? void 0 : _a.handleDocumentSymbolRequest(params)) !== null && _b !== void 0 ? _b : []);
    }
    async handleWorkspaceSymbolRequest(params) {
        return Promise.all(Array.from(this._processors.values()).flatMap(processor => processor.handleWorkspaceSymbolRequest(params))).then(symbolsList => symbolsList.flat());
    }
}
//# sourceMappingURL=MessageProcessor.js.map